#include <iostream>
#include <fstream>
#include <string.h>
#include <string>
#include <list>
#include <map>
#include <File.h>
#include <stdio.h>
#include <Directory.h>
#include <Path.h>
#include <NodeInfo.h>
#include <time.h>
#include <vector>
#include <algorithm>
#include <strstream>

#include "libpq++.h"

#include "utils.h"

//#define DEBUG 

PgDatabase		data("dbname=beindexed host=127.0.0.1");

/**
	A list of 'stop words', populated by init_stop_words()
*/
vector<string>	g_stop_words;

const char * _stop_words[] = {
// english
	"the", "of", "and", "to", "in", "is", "you", "that", "it", "he" "for",
	"was", "on", "are", "as", "with", "his", "they", "at", "be", "this",
	"from", "have", "or", "by", "one", "had", "not", "but", "what", "all", 
	"were", "when", "we", "there", "can", "an", "your", "which", "their", "said",
	"if", "do", "will", "each", "about", "how", "up", "out", "them", "then", "she",
	"many", "some", "so", "these", "would", "other", "into", "has", "more", "her", 
	"two", "like", "him", "see", "time", "could", "no", "make", "than", "first",
	"been", "its", "who", "now", "people", "my", "made", "over", "did", "down",
	"only", "way", "find", "use", "may", "water", "long", "little", "very", 
	"after", "words", "called", "just", "where", "most", "know", "get",
	"through", "back", "much", "before", "go", "good", "new", "write", "our",
/*
	used me man too any day same right look think also around another came come work three word must because does part even place well such here take why things help put years different away again off went old number great tell men say small every found still between name should Mr home big give air line set own under read last never us left end along while might next sound below saw something thought both few those always looked show large often together asked house don't world going want

school important until 1 form food keep children feet land side without boy once animals life enough took sometimes four head above kind began almost live page got earth need far hand high year mother light parts country father let night following 2 picture being study second eyes soon times story boys since white days ever paper hard near sentence better best across during today others however sure means knew it's try told young miles sun ways thing whole hear example heard several change answer room sea against top turned 3 learn point city play toward five using himself usually 
*/

// svenska
// from http://spraakdata.gu.se/lb/statistik/lbgrafsv97.html
	"och", "att", "det", "som", "en", "på", "av", "är", "för", "med", "till",
	"den", "har", "de", "inte", "om", "ett", "han", "men", "var", "jag", "sig", "vi", 

// end of list
	""
};


/**
	Perform a query using the global PgDatabase instance.
*/
int
do_query(const char * query )
{
	int status;
	
	#ifdef DEBUG
	cout << query << endl;
	#endif
	
	status = data.Exec( query );
	
	switch ( status )
	{
		case PGRES_TUPLES_OK:
			break;
		case PGRES_COMMAND_OK:
			break;
		default:
			cout << "SQL Error: " << data.ErrorMessage() << endl;
			return -1;
	}
	
	return 0;
}


/**
	Set up g_stop_words, a list of 'stop words'.
	
	Will eventually get the list of words to ignore from the database so the
	user can add or remove words as needed.
*/
void
init_stop_words()
{
	for ( int i=0; string(_stop_words[i]) != ""; i++ )
	{
		g_stop_words.push_back( string(_stop_words[i]) );
	}
}

/**
	Determine if string is a 'stop word', a very common word that it makes no sense
	to include in the database, like 'the' and 'or'.
	
	init_stop_words() must be called before using this function
*/
bool
is_stop_word( string word )
{
	return find( g_stop_words.begin(), g_stop_words.end(), word ) != g_stop_words.end();
}

/**
	Index a text/plain-file.
*/
void 
text_plain_index_file( const char * filename, map<string,int> & wordlist )
{
	ifstream i(filename);
	
	string word;
	
	map<string,int>::iterator iter;
	
	while ( !i.eof() )
	{
		skip_non_text(i);
		word = toLower( get_word(i) );
		
		if ( word == "" )
			// we get this at eof
			continue;
		
		if ( word.length() < MIN_WORD_LENGTH || word.length() > MAX_WORD_LENGTH )
			// skip too small and too large words
			continue;
		
		if ( is_number(word) )
			// skip numbers
			continue;
		
		if ( is_stop_word(word) )
			// skip stop words
			continue;
		
		iter = wordlist.find( word );
		if ( iter != wordlist.end() )
		{
			wordlist[word] = iter->second + 1;
		} else 
		{
			wordlist[word] = 1;
		}
	}
}


/**
	Get the MIME-type for a given file.
	
	Will eventually throw an error if the operation fails
*/
string
get_file_type( const char * filename )
{
	BEntry entry(filename);
	BNode node(&entry);
	BNodeInfo info(&node);
		
	char type[256];
	
	if ( info.GetType(type) != B_OK )
		// throw error here
		return "";
	else
		return type;
}

/**
	Defines a type-handler relationship, used by init_type_handlers() and add-ons.
*/
typedef struct type_handler_t
{
	string type;
	void (*handler)(const char *, map<string,int> &);
};

/**
	Vector containing a list of current type handlers, use init_type_handlers() to
	set it up before calling index_file()
*/
vector<type_handler_t> g_type_handlers;

/**
	Initialize g_type_handlers, must be called before any call to index_file(), but
	must only be called once!
	
	Will eventually load add-ons and add them to the list as well to provide an easy
	way to add support for more formats.
*/
void
init_type_handlers()
{
	g_type_handlers.push_back( (type_handler_t){"text/plain",text_plain_index_file} );
	g_type_handlers.push_back( (type_handler_t){"text/x-source-code",text_plain_index_file} );
}


/**
	Check the file's type and call the corresponding handler.
	Will eventually throw an error if no handler is found, or an error occurs.
	Don't forget to call init_type_handlers() once at the start of the program
	to set up type > handler relationships (and load any present add-ons)
*/
void
index_file( const char * filename, map<string,int> & wordlist )
{
	string type = get_file_type(filename);
	
	for ( vector<type_handler_t>::iterator i = g_type_handlers.begin(); i != g_type_handlers.end(); i++ )
	{
		if ( (*i).type == type )
		{
			(*i).handler(filename,wordlist);
			return;
		}
	}
}

/**
	Query the database for the fileTable.id corresponding to path
*/
int
get_file_id( string path )
{
	string query = "SELECT id FROM fileTable WHERE path='";
	query += path;
	query += "';";
	do_query( query.c_str() );
	
	if ( data.Tuples() == 0 )
		return -1;
	
	return atoi( data.GetValue(0,0) );
}

/**
	Query the database for the wordTable.id corresponding to word
*/
int
get_word_id( string word )
{
	string query = "SELECT id FROM wordTable WHERE word='";
	query += word;
	query += "';";
	do_query( query.c_str() );
	
	if ( data.Tuples() == 0 )
		return -1;
	
	return atoi( data.GetValue(0,0) );
}

/**
	Checks if a handler that can handle the file's type is present.
	
	@param path Path of file to be examined
*/
bool
is_handler_present( const char * path )
{
	string type = get_file_type(path);
	
	if ( type == "" )
		return false;
	
	for ( vector<type_handler_t>::iterator i = g_type_handlers.begin(); i != g_type_handlers.end(); i++ )
	{
		if ( (*i).type == type )
		{
			return true;
		}
	}
	
	return false;
}

/**
	Checks if a file should be ignored and not included in the database.
	This is accomplished by checking a number of things:
	
		* If an attribute, BeIndexed:Rule, is set to 'ignore' or not.
		--- The rest are only checked if just_attr is false
		* If the file is a symlink
		* If there is an appropriate handler (unless check_handler is false)
	
	@param just_attr Determines if other checks than the attribute one should be made, default false
*/
bool 
should_ignore( const char * path, bool just_attr=false )
{
	BEntry entry(path);
	BNode node(&entry);
	
	char rule[256];
	
	ssize_t result = node.ReadAttr( "BeIndexed:Rule", B_STRING_TYPE, 0, rule, sizeof(rule) );
	
	if ( result > 0 )
	{ // attribute present, check it!
		rule[result] = 0;
		
		if ( strcmp(rule,"ignore") == 0 )
		{
			//#ifdef DEBUG
			cout << "BeIndexed:Rule = ignore: " << path << endl;
			//#endif
			return true;
		}
	}
	
	if ( just_attr )
		return false;
	
	if ( entry.IsSymLink() )
		return true;
	
	if ( !is_handler_present(path) )
		return true;
	
	return false;
}

/**
	Make sure that all words in list are present in wordTable
	
	@param wordlist List of words (and their weights) to be added to database
*/
void
add_words_to_wordtable( map<string,int> & wordlist )
{
	string query;
	
	map<string,int>::iterator iter = wordlist.begin();
	
	while ( iter != wordlist.end() )
	{
		int id = get_word_id(iter->first);
		
		if ( id < 0 )
		{ // add word
			query = "INSERT INTO wordTable (word) VALUES ('";
			query += iter->first;
			query += "')";
			
			if ( do_query(query.c_str()) < 0 )
			{	// something wen't bad. return!
				return;
			}
		}
		
		iter++;
	}
}

/**
	Get the filesystem modification date for a file
*/
string
get_modification_date( const char * path )
{
	time_t	mod_time;
		
	BEntry entry(path);
		
	entry.GetModificationTime(&mod_time);
		
	tm *	mod_tm;
	mod_tm = localtime(&mod_time);
		
	// show what we're working with..
	char mod_date_string[100];
	sprintf(mod_date_string, "%d-%.02d-%.02d %.02d:%.02d:%.02d",
		mod_tm->tm_year + 1900,
		mod_tm->tm_mon + 1,
		mod_tm->tm_mday,
		mod_tm->tm_hour,
		mod_tm->tm_min,
		mod_tm->tm_sec
	);
	
	return mod_date_string;
}

// forward declaration..
void add_dir_to_database( const char * );

/**
	Indexes a file and adds the result to the database.
	
	@param path Path to the file to add
*/
void
add_file_to_database( const char * path )
{
	BEntry entry(path);
	
	if ( entry.IsDirectory() )
	{ // ooh, it's a dir and not a file!
		add_dir_to_database( path );
		return;
	}
	
	// update mime info so we know what we're handling
	update_mime_info(path,false/* recursive */,true /* sync */,false /* force */);
	
	// see if this is a file we can and want to handle
	if ( should_ignore(path) )
		return;
	
	char number[64];
	
	map<string,int> wordlist;
	
//	#ifdef DEBUG
	// seem like we want to index this, carry on..
	cout << "indexing file '" << path << "'" << endl;
//	#endif
	
	// look for file id in fileTable
	int file_id = get_file_id(path);
	
	if ( file_id < 0 )
	{ // file not in fileTable, add it
		string query = "INSERT INTO fileTable (path) VALUES ('";
		query += path;
		query += "');";
		do_query( query.c_str() );
		
		file_id = get_file_id(path);
	} else
	{ // check if file has been modified since it was indexed
		// get indexed_at date from database
		sprintf(number,"%d",file_id);
		string query = string("SELECT * FROM fileTable WHERE id=") + string(number);
		
		do_query( query.c_str() );
		
		string index_date = data.GetValue(0,"indexed_at");
		index_date.resize(19);
		
		// get modification date from filesystem
		string fs_mod_date = get_modification_date(path);
		
		#ifdef DEBUG
		cout << "  i: " << index_date << " vs m: " << fs_mod_date << endl;
		#endif
		
		if ( index_date >= fs_mod_date )
		{ // no need to index, already up to date
//			#ifdef DEBUG
			cout << "  no need to index" << endl;
//			#endif
			return;
		}
	}
	
	// analyze the database once in a while to speed things up
	static int added_files = 0;
	
	if ( added_files++ % 50 == 0 )
	{
		cout << "Analyzing database.." << endl;
		do_query("ANALYZE;");
	}
	
	// index file
	index_file(path,wordlist);
	
	#ifdef DEBUG
	cout << "  found " << wordlist.size() << " words" << endl;
	#endif
	
	// add words to wordtable
	string query;
	
	add_words_to_wordtable(wordlist);
		
	// get file size so we can use it to calculate weights
	BFile file(path,B_READ_ONLY);
	off_t _size=0;
	file.GetSize(&_size);
	double filesize = _size;
	
	// begin update
	
	do_query("BEGIN;");
		
		// delete old words
		sprintf(number,"%d",file_id);
		query = "DELETE FROM contentTable WHERE file=";
		query += number;
		query += ";";
		
		do_query( query.c_str() );
		
		// add new content
		for ( map<string,int>::iterator iter=wordlist.begin(); iter != wordlist.end(); iter++ )
		{
			query = "INSERT INTO contentTable (word,weight,file) VALUES ('";
			sprintf(number,"%d",get_word_id(iter->first));
			query += number;
			query += "',";
			sprintf(number,"%f",
				(iter->second*strlen(iter->first.c_str()))
				/
				filesize
			);
			query += number;
			query += ",";
			sprintf(number,"%d",file_id);
			query += number;
			query += ");";
			
			if ( do_query(query.c_str() ) < 0 )
			{ // error adding, do an ABORT and return
				do_query("ABORT;");
				// actually we should throw an error, but..
				return;
			}
		}
	
	sprintf(number,"%d",file_id);
	query = "UPDATE fileTable SET indexed_at=now() WHERE id=";
	query += number;
	query += ";";
	
	do_query(query.c_str());
	
	do_query("COMMIT;");
}

/**
	Iterates through the files in a directory, adding them all to the database.
*/
void
add_dir_to_database( const char * path )
{
	BDirectory dir(path);
	BEntry entry;
	BPath p;
	
	if ( should_ignore(path,true /* only check the attribut since it's a dir*/ ) )
		return;
	
	while ( dir.GetNextEntry(&entry) == B_OK )
	{
		entry.GetPath(&p);
		
		add_file_to_database(p.Path());
	}
}

/**
	For each file: make sure that the file still exists in the filesystem, and if
	it doesn't remove it from the database.
*/
void
purge_deleted_files()
{
	do_query("SELECT * FROM fileTable");
	
	// a list of filetable.id to delete
	list<string> deleted_files;
	
	for ( int i=0; i<data.Tuples(); i++ )
	{
		string path = data.GetValue(i,"path");
		
		BEntry entry( path.c_str() );
		
		if ( entry.Exists() )
			continue;
		
		// file deleted, mark for deletion
		deleted_files.push_back( data.GetValue(i,"id") );
	}
	
	list<string>::iterator i;
	
	for ( i=deleted_files.begin(); i != deleted_files.end(); i++ )
	{
		char query[512];
		
		{ // delete contentTable references
			strstream str;
			str << "DELETE FROM contentTable WHERE file=" << *i;
		
			str.getline(query,sizeof(query)-1);
		
			do_query( query );
		}
		{ // delete fileTable reference
			strstream str;
			str << "DELETE FROM fileTable WHERE id=" << *i;
		
			str.getline(query,sizeof(query)-1);
		
			do_query( query );
		}
	}
}

/**
	Plain ol' main(). Performs some init, then adds something to the database
*/
int 
main(int numArg, char ** argv)
{
	init_stop_words();
	init_type_handlers();
	
	if ( numArg < 2 )
	{
		cout << "usage: Indexer <file or directory to index>" << endl;
	} else {
		cout << "Removing deleted files from index..." << endl;
		purge_deleted_files();
		
		add_file_to_database(argv[1]);
	}
	
	return 0;
}
